!{\src2tex{textfont=tt}}
!!****m* ABINIT/abi_libxc_functionals
!! NAME
!!  abi_libxc_functionals
!!
!! FUNCTION
!!  Module containing interfaces to the LibXC library, for exchange
!!  correlation potentials and energies. The interfacing between
!!  the ABINIT and LibXC formats and datastructures happens here.
!!  Also contains basic container datatype for LibXC interfacing.
!!
!! COPYRIGHT
!! Copyright (C) 2008-2010 ABINIT group (MOliveira,MT)
!! This file is distributed under the terms of the
!! GNU General Public License, see ~abinit/COPYING
!! or http://www.gnu.org/copyleft/gpl.txt .
!!
!! SOURCE

#if defined HAVE_CONFIG_H
#include "config.h"
#endif

module abi_libxc_functionals

  use abi_defs_basis
  use abi_interfaces_lowlevel
#if defined HAVE_LIBXC
  use xc_f90_types_m
  use libxc_funcs_m
  use xc_f90_lib_m
#endif

  implicit none

#if defined HAVE_LIBXC
  type abi_libxc_functional
    private
    integer         :: family ! LDA, GGA, etc.
    integer         :: id     ! identifier
    integer         :: nspin  ! # of spin components
    type(xc_f90_pointer_t) :: conf ! the pointer used to call the library
    type(xc_f90_pointer_t) :: info ! information about the functional
  end type abi_libxc_functional

  type(abi_libxc_functional) :: abi_funcs(2)

  private
  public :: abi_libxc_functionals_getvxc
  public :: abi_libxc_functionals_init
  public :: abi_libxc_functionals_end
  public :: abi_libxc_functionals_isgga
  public :: abi_libxc_functionals_ismgga
#endif

contains
!!***

#if defined HAVE_LIBXC

!--------------------------------------------------------------------------------------

!!****f* abi_libxc_functionals/abi_libxc_functionals_init
!! NAME
!!  abi_libxc_functionals_init
!!
!! FUNCTION
!!  Initialize the desired XC functional, from LibXC.
!!  * Call the LibXC initializer
!!  * Fill preliminary fields in module structures.
!!
!! SOURCE

  subroutine abi_libxc_functionals_init(ixc,nspden)

    implicit none

!Arguments ------------------------------------
    integer, intent(in) :: nspden
    integer, intent(in) :: ixc
!Local variables-------------------------------
    integer :: i, ii
    character(len=500) :: message
    type(xc_f90_pointer_t) :: str

! *************************************************************************

    abi_funcs(1)%id = -ixc/1000
    abi_funcs(2)%id = -ixc - abi_funcs(1)%id*1000

    abi_funcs(1)%nspin = nspden
    abi_funcs(2)%nspin = nspden

    do i = 1, 2
      if (abi_funcs(i)%id == 0) then
        abi_funcs(i)%family = 0
        cycle
      end if

      ! Get XC functional family
      abi_funcs(i)%family = xc_f90_family_from_id(abi_funcs(i)%id)
      select case (abi_funcs(i)%family)
      case (XC_FAMILY_LDA, XC_FAMILY_GGA, XC_FAMILY_MGGA)
        call xc_f90_func_init(abi_funcs(i)%conf,abi_funcs(i)%info,abi_funcs(i)%id,nspden)
      case default
        write(message, '(4a,i8,2a,i8,6a)' )ch10,&
             &    ' abi_libxc_functionals_init : ERROR -',ch10,&
             &    '  Invalid IXC = ',ixc,ch10,&
             &    '  The LibXC functional family ',abi_funcs(i)%family,&
             &    '  is currently unsupported by ABINIT',ch10,&
             &    '  (-1 means the family is unknown to the LibXC itself)',ch10,&
             &    '  Please consult the LibXC documentation',ch10
        call abi_wrtout(std_out,message,'COLL')
        call abi_leave_new('COLL')
      end select

      if (abi_funcs(i)%id == XC_LDA_C_XALPHA) then
        call xc_f90_lda_c_xalpha_set_par(abi_funcs(i)%conf,zero)
      end if
    end do

  end subroutine abi_libxc_functionals_init
!!***

!--------------------------------------------------------------------------------------

!!****f* abi_libxc_functionals/abi_libxc_functionals_end
!! NAME
!!  abi_libxc_functionals_end
!!
!! FUNCTION
!!  End usage of LibXC functional. Call LibXC end func. and deallocate module contents.
!!
!!
!! SOURCE
  subroutine abi_libxc_functionals_end()

    implicit none
    integer :: i

    do i = 1, 2
      if (abi_funcs(i)%id == 0) cycle
      call xc_f90_func_end(abi_funcs(i)%conf)
    end do

  end subroutine abi_libxc_functionals_end
!!***

!--------------------------------------------------------------------------------------

!!****f* abi_libxc_functionals/abi_libxc_functionals_isgga
!! NAME
!!  abi_libxc_functionals_isgga
!!
!! FUNCTION
!!  Test function to identify whether the presently used functional is a GGA or not
!!
!! SOURCE
  function abi_libxc_functionals_isgga()

    implicit none
    logical :: abi_libxc_functionals_isgga

! *************************************************************************

    abi_libxc_functionals_isgga =(any(abi_funcs%family == XC_FAMILY_GGA))

  end function abi_libxc_functionals_isgga
!!***

!--------------------------------------------------------------------------------------

!!****f* abi_libxc_functionals/abi_libxc_functionals_ismgga
!! NAME
!!  abi_libxc_functionals_ismgga
!!
!! FUNCTION
!!  Test function to identify whether the presently used functional is a Meta-GGA or not
!!
!! SOURCE
  function abi_libxc_functionals_ismgga()

    implicit none
    logical :: abi_libxc_functionals_ismgga

! *************************************************************************

    abi_libxc_functionals_ismgga =(any(abi_funcs%family == XC_FAMILY_MGGA))

  end function abi_libxc_functionals_ismgga
!!***

!--------------------------------------------------------------------------------------

!!****f* abi_libxc_functionals/abi_libxc_functionals_getvxc
!! NAME
!!  abi_libxc_functionals_getvxc
!!
!! FUNCTION
!!  Return XC potential and energy, from input density (event gradient etc...)
!!
!! SOURCE

  subroutine abi_libxc_functionals_getvxc(ndvxc,nd2vxc,npts,nspden,order,rho,exc,vxc,grho2,vxcgr, &
&                                         lrho,vxclrho,tau,vxctau,dvxc,d2vxc)

    implicit none

!Arguments ------------------------------------
    integer, intent(in) :: ndvxc,nd2vxc,npts,nspden,order
    real(dp),intent(in)  :: rho(npts,nspden)
    real(dp),intent(out) :: vxc(npts,nspden), exc(npts)
    real(dp),intent(in),optional :: grho2(npts,2*min(nspden,2)-1)
    real(dp),intent(out),optional :: vxcgr(npts,3)
    real(dp),intent(in),optional :: lrho(npts,nspden)
    real(dp),intent(out),optional :: vxclrho(npts,nspden)
    real(dp),intent(in),optional :: tau(npts,nspden)
    real(dp),intent(out),optional :: vxctau(npts,nspden)
    real(dp),intent(out),optional :: dvxc(npts,ndvxc)
    real(dp),intent(out),optional :: d2vxc(npts,nd2vxc)
!Local variables-------------------------------
    integer  :: i, ipts
    real(dp) :: c, rhotmp(nspden), exctmp, sigma(3), vsigma(3), vxctmp(nspden)
    real(dp) :: v2rho2(3),v2rhosigma(6),v2sigma2(6),v3rho3(4)
    real(dp) :: lrhotmp(nspden), tautmp(nspden), vxclrhotmp(nspden), vxctautmp(nspden)
    real(dp), allocatable :: gnon(:)
    character(len=500) :: message

! *************************************************************************

    ! Inititalize all relevant arrays to zero
    vxc=zero ; exc=zero ; vxctmp=zero ; exctmp=zero
    v2rho2=zero ; v2rhosigma=zero ; v2sigma2=zero
    v3rho3=zero
    if (order**2 >1) dvxc=zero
    if (order**2 >4) d2vxc=zero
    if (any(abi_funcs%family == XC_FAMILY_GGA)) vxcgr=zero
    if (any(abi_funcs%family == XC_FAMILY_MGGA)) then
      vxcgr=zero ; vxclrho=zero ; vxctau=zero
    end if

    !The TB09 MGGA functional requires an extra quantity
    if (any(abi_funcs%id == XC_MGGA_X_TB09)) then
      allocate(gnon(npts))
      do ipts = 1, npts
        if (sum(rho(ipts, :)) <= 1e-7_dp) then
          gnon(ipts) = zero
        else
          if (nspden == 1) then
            gnon(ipts) = sqrt(grho2(ipts,1))/rho(ipts, 1)
          else
            gnon(ipts) = sqrt(grho2(ipts,3))/sum(rho(ipts, :))
          end if
        end if
      end do
      c = -0.012_dp + 1.023_dp*sqrt(sum(gnon)/npts)
      do i = 1, 2
        if (abi_funcs(i)%id == XC_MGGA_X_TB09) then
          call xc_f90_mgga_x_tb09_set_par(abi_funcs(i)%conf, c)
        end if
      end do
      deallocate(gnon)
    end if

    !Loop over points
    do ipts = 1, npts

      ! Convert the quantities provided by ABINIT to the ones needed by libxc
      if (nspden == 1) then
        ! ABINIT passes rho_up in the spin-unpolarized case, while the libxc
        ! expects the total density
        rhotmp(1:nspden) = two*rho(ipts,1:nspden)
      else
        rhotmp(1:nspden) = rho(ipts,1:nspden)
      end if
      if (any(abi_funcs%family == XC_FAMILY_GGA) .or. any(abi_funcs%family == XC_FAMILY_MGGA)) then
        sigma=zero
        if (nspden==1) then
          ! ABINIT passes |grho_up|^2 while Libxc needs |grho_tot|^2
          sigma(1) = four*grho2(ipts,1)
        else
          ! ABINIT passes |grho_up|^2, |grho_dn|^2, and |grho_tot|^2
          ! while Libxc needs |grho_up|^2, grho_up.grho_dn, and |grho_dn|^2
          sigma(1) = grho2(ipts,1)
          sigma(2) = (grho2(ipts,3) - grho2(ipts,1) - grho2(ipts,2))/two
          sigma(3) = grho2(ipts,2)
        end if
      end if
      if (any(abi_funcs%family == XC_FAMILY_MGGA)) then
        if (nspden==1) then
          lrhotmp(1:nspden) = two*lrho(ipts,1:nspden)
          tautmp(1:nspden) = four*tau(ipts,1:nspden)
        else
          lrhotmp(1:nspden) = lrho(ipts,1:nspden)
          tautmp(1:nspden) = two*tau(ipts,1:nspden)
        end if
      end if

      !Loop over functionals
      do i = 1,2
        if (abi_funcs(i)%id == 0) cycle

        !Get the potential (and possibly the energy)
        if (iand(xc_f90_info_flags(abi_funcs(i)%info), XC_FLAGS_HAVE_EXC) .ne. 0) then
          select case (abi_funcs(i)%family)
          case (XC_FAMILY_LDA)
            call xc_f90_lda_exc_vxc(abi_funcs(i)%conf,1,rhotmp(1),exctmp,vxctmp(1))
            if (order**2 > 1) then
              call xc_f90_lda_fxc(abi_funcs(i)%conf,1,rhotmp(1),v2rho2(1))
            endif
            if (order**2 > 4) then
              call xc_f90_lda_kxc(abi_funcs(i)%conf,1,rhotmp(1),v3rho3(1))
            endif
          case (XC_FAMILY_GGA)
            call xc_f90_gga_exc_vxc(abi_funcs(i)%conf,1,rhotmp(1),sigma(1),exctmp,vxctmp(1),vsigma(1))
            if (order**2 > 1) then
              call xc_f90_gga_fxc(abi_funcs(i)%conf,1,rhotmp(1),sigma(1),v2rho2(1),v2rhosigma(1),v2sigma2(1))
            endif
          case (XC_FAMILY_MGGA)
            call xc_f90_mgga_exc_vxc(abi_funcs(i)%conf,1,rhotmp(1),sigma(1),lrhotmp(1),&
                        tautmp(1),exctmp,vxctmp(1),vsigma(1),vxclrhotmp(1),vxctautmp(1))
          end select

        else
          exctmp=zero
          select case (abi_funcs(i)%family)
          case (XC_FAMILY_LDA)
            call xc_f90_lda_vxc(abi_funcs(i)%conf,1,rhotmp(1),vxctmp(1))
          case (XC_FAMILY_GGA)
            call xc_f90_gga_vxc(abi_funcs(i)%conf,1,rhotmp(1),sigma(1),vxctmp(1),vsigma(1))
          case (XC_FAMILY_MGGA)
            call xc_f90_mgga_vxc(abi_funcs(i)%conf,1,rhotmp(1),sigma(1),lrhotmp(1),&
                        tautmp(1),vxctmp(1),vsigma(1),vxclrhotmp(1),vxctautmp(1))
          end select
        end if

        exc(ipts) = exc(ipts) + exctmp
        vxc(ipts,1:nspden) = vxc(ipts,1:nspden) + vxctmp(1:nspden)

        if (order**2>1) then
          select case (abi_funcs(i)%family)
          case (XC_FAMILY_LDA)
            if (nspden==1) then
              if(order>=2) then
                dvxc(ipts,1)=dvxc(ipts,1)+v2rho2(1)
                if(order==3) then
                  d2vxc(ipts,1)=d2vxc(ipts,1)+v3rho3(1)
                endif
              else
                dvxc(ipts,1)=dvxc(ipts,1)+v2rho2(1)
                dvxc(ipts,2)=dvxc(ipts,2)+v2rho2(1)
              endif
            else
              dvxc(ipts,1)=dvxc(ipts,1)+v2rho2(1)
              dvxc(ipts,2)=dvxc(ipts,2)+v2rho2(2)
              dvxc(ipts,3)=dvxc(ipts,3)+v2rho2(3)
              if(order==3) then
                d2vxc(ipts,1)=d2vxc(ipts,1)+v3rho3(1)
                d2vxc(ipts,2)=d2vxc(ipts,2)+v3rho3(2)
                d2vxc(ipts,3)=d2vxc(ipts,3)+v3rho3(3)
                d2vxc(ipts,4)=d2vxc(ipts,4)+v3rho3(4)
              endif
            endif
          case (XC_FAMILY_GGA)
            if (i==1) then
              if (nspden==1) then
                dvxc(ipts,1)=v2rho2(1)*two
                dvxc(ipts,2)=dvxc(ipts,1)
                dvxc(ipts,3)=two*two*vsigma(1)
                dvxc(ipts,4)=dvxc(ipts,3)
                dvxc(ipts,5)=four*two*v2rhosigma(1)
                dvxc(ipts,6)=dvxc(ipts,5)
                dvxc(ipts,7)=two*four*four*v2sigma2(1)
                dvxc(ipts,8)=dvxc(ipts,7)
              else
                dvxc(ipts,1)=v2rho2(1)
                dvxc(ipts,2)=v2rho2(3)
                dvxc(ipts,3)=two*vsigma(1)
                dvxc(ipts,4)=two*vsigma(3)
                dvxc(ipts,5)=two*v2rhosigma(1)
                dvxc(ipts,6)=two*v2rhosigma(6)
                dvxc(ipts,7)=four*v2sigma2(1)
                dvxc(ipts,8)=four*v2sigma2(6)
              end if
            else
              if (nspden==1) then
                dvxc(ipts,9)=v2rho2(1)
                dvxc(ipts,10)=dvxc(ipts,9)
                dvxc(ipts,11)=dvxc(ipts,9)
                dvxc(ipts,12)=two*vsigma(1)
                dvxc(ipts,13)=two*v2rhosigma(1)
                dvxc(ipts,14)=dvxc(ipts,13)
                dvxc(ipts,15)=four*v2sigma2(1)
              else
                dvxc(ipts,9)=v2rho2(1)
                dvxc(ipts,10)=v2rho2(2)
                dvxc(ipts,11)=v2rho2(3)
                dvxc(ipts,12)=two*vsigma(1)
                dvxc(ipts,13)=two*v2rhosigma(1)
                dvxc(ipts,14)=two*v2rhosigma(6)
                dvxc(ipts,15)=four*v2sigma2(1)
              end if
            end if
          end select
        end if

        if (any(abi_funcs%family == XC_FAMILY_GGA) .or. any(abi_funcs%family == XC_FAMILY_MGGA)) then
          !Convert the quantities returned by Libxc to the ones needed by ABINIT
          if (nspden == 1) then
            vxcgr(ipts,3) = vxcgr(ipts,3) + vsigma(1)*two
          else
            vxcgr(ipts,1) = vxcgr(ipts,1) + two*vsigma(1) - vsigma(2)
            vxcgr(ipts,2) = vxcgr(ipts,2) + two*vsigma(3) - vsigma(2)
            vxcgr(ipts,3) = vxcgr(ipts,3) + vsigma(2)
          end if
        end if
        if (any(abi_funcs%family == XC_FAMILY_MGGA)) then
          vxclrho(ipts,1:nspden) = vxclrho(ipts,1:nspden) + vxclrhotmp(1:nspden)
          vxctau(ipts,1:nspden) = vxctau(ipts,1:nspden) + two*vxctautmp(1:nspden)
        end if

      end do
    end do

  end subroutine abi_libxc_functionals_getvxc

#endif

end module
!!***
